// lighting.c

/*  Copyright (C) 1996-1997  Id Software, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

    See file, 'COPYING', for details.
*/

#include "light.h"
// #include "threads.h"	// gb

#define LOGFILENAME ".light.log"	// gb

/*

NOTES
-----

*/

float		scaledist = -1;	      // "-dist" value
float		scalecos = 0.5;	      // Default angle sensitivity
float		ShadowSense = -1;     // Angle (shadow) sensitivity for 2nd sunlight
float		rangescale = -1;      // "-range" value
float		GateVal = 0;	      // Fade Gate value (limit attenuated lights)
int	        worldminlight = -1;
int	        worldmaxlight = -1;

byte		*filebase, *file_p, *file_end;

dmodel_t	*bspmodel;
int		bspfileface;	      // next surface to dispatch

vec3_t		bsp_origin;
vec3_t		*faceoffset;	      // Rotating entities related

int		OverSample = 1;	      // 1 = No oversampling
int		SoftLight = -1;	      // -1 = Disabled
int		SoftDist = SOFTDIST;  // Distance tolerance for lights behind surface
int		SingleDist;	      // Distance tolerance for lights behind surface in SingleLightFace
int		SkyDist;	      // Distance tolerance for lights behind surface in SkyLightFace
int		LightCap = 0;	      // 0 = Disabled
int		NumSurfPts = 0;	      // Total # surface points
unsigned int	FastLight = 0;	      // 0 = No fast lighting
qboolean	NoLight = false;      // Disables all light entities except global
qboolean	SrcLight = false;     // Disables all unsourced light entities
qboolean	NoWarnings = false;   // Disable repetitive warnings
qboolean	DisableScreen = false;
qboolean	OldLight = false;     // Disables wait/delay/mangle
qboolean	TyrLiteMode = false;  // Enables TyrLite mode
qboolean	TyrLite95Mode = false;// Enables TyrLite95 mode
qboolean	ArghLiteMode = false; // Enables ArghLite mode
qboolean	AddMinLight = false;  // Enables additive minlight
qboolean	UnsupDetails = false; // Prints out unsupported details
qboolean	DetectKeys = false;   // Prints out detected keys summary
qboolean	NoSkillChk = false;   // Disable fuzzy skill target check
qboolean	NoFlash = false;      // Disables flashing lights
qboolean	GlobRange = false;    // Enables global light range
qboolean	IKLiteMode = false;   // Enables IKLite mode
qboolean	IKAngle = false;      // Enables IKLite angle sensitivity
qboolean	LightDLXMode = false; // Enables LightDLX mode
qboolean	OnlyEnts = false;     // Only process switchable lights
qboolean	NoWrite = false;      // Disable bsp write
qboolean	NoReverse = false;    // Prevents reverse entity key order
qboolean	GenCompatible = false;// Enables closer general compatibility
qboolean	ATCompatible = false; // Enables closer Argh/TyrLite compatibility
qboolean	TyrCompatible = false;// Enables closer TyrLite compatibility
qboolean	EnhancedTP = false;   // Enables QuArK/Hammer Enhanced Texture Positioning support
qboolean	AntiLights = false;   // True if negative lights used
qboolean	NoAnti = false;	      // True if negative lights disabled
qboolean	NoAntiOption = false; // True if negative lights disabled via "-noanti" option
qboolean	KinnDelay = false;    // Translates delay 2 lights into delay 5
qboolean	SolidSky = false;     // True if also solid sky brushes will cast sunlight
qboolean	PreScan;	      // True when only calculating # total surfpts, avoid ray tracing

int		SunLight[2] = {-1, -1};
int		SunMangleVal[3] = {-1, -1, -1};
int		NoOfSuns, NoOfHSuns;
vec3_t		SunMangle[NOOFSUNS];
vec_t		DistFactor1;	      // IKLite related values
vec_t		DistFactor2;

double		start = 0, end;
FILE		*logfile;

void logvprintf (char *fmt, va_list argptr)
{
	if (PreScan)
		return; // No printouts during prescan

	if (!NoWarnings || !DisableScreen)
	{
		ShowPercent (NULL, 0, -1);

		vprintf (fmt, argptr);
		fflush (stdout);
	}

        vfprintf (logfile, fmt, argptr);
        fflush (logfile);
}

void logprintf (char *fmt, ...)
{
	va_list argptr;

	va_start (argptr, fmt);
	logvprintf (fmt,argptr);
	va_end (argptr);
}

void logwprintf (char *fmt, ...)
{
	va_list argptr;

	DisableScreen = true;
	va_start (argptr, fmt);
	logvprintf (fmt,argptr);
	va_end (argptr);
	DisableScreen = false;
}

void PrintFinish (void)
{
	char Str[20];

	if (start != 0)
	{
		end = I_FloatTime ();

		SecToStr ((time_t)(end - start + 0.5), Str, true);

		logprintf ("\nElapsed time : %s\n", Str);
	}

	fclose (logfile);
}

void ErrorExit (void)
{
	PrintFinish ();
	exit (1);
}

/*
==================
ToRad
Converts from degrees to radians
==================
*/
vec_t ToRad (vec_t Degree)
{
	return Degree * Q_PI / 180;
}

/*
==================
ToDegree
Converts from radians to degrees
==================
*/
vec_t ToDegree (vec_t Rad)
{
	return Rad * 180 / Q_PI;
}

byte *GetFileSpace (int size)
{
	byte *buf;

	LOCK;
	buf = file_p;

	file_p = (byte *)(((long)file_p + 3)&~3);

	if (file_p - buf > 0)
		memset (buf, 0, file_p - buf); // Clear pad bytes

	buf = file_p;
	file_p += size;
	UNLOCK;
	if (file_p > file_end)
		Error ("Light data size exceeded, max = %s", PrtSize(MAX_MAP_LIGHTING));
	return buf;
}


void LightThread (void *junk)
{
	int			i;

	if (NewLine)
		fprintf (logfile, "\n");
	else
		logprintf ("\n");

	if (SimpPercent)
		printf ("\n");

	ShowPercent ("Light", 0, 0);

	while (1)
	{
		LOCK;
		i = bspfileface++;
		UNLOCK;

		if (i >= numfaces)
		{
			ShowPercent (NULL, 100, 100);

			if (!SimpPercent)
				printf ("\n");

			printf ("\n");

			return;
		}

		ShowPercent (NULL, i, numfaces);

		LightFace (i, faceoffset[i]);
	}
}

void FindFaceOffsets (void)
{
	int	 i, j;
	entity_t *ent;
	char	 name[20];
	vec3_t	 org;

	faceoffset = malloc (numfaces * sizeof(vec3_t));
	memset (faceoffset, 0, numfaces * sizeof(vec3_t));

	for( i = 1; i < nummodels; i++ )
	{
		sprintf (name, "*%d", i);
		ent = FindEntityWithKeyPair ("model", name);

		if (!ent)
			logprintf ("WARNING: FindFaceOffsets: Couldn't find entity for model %s\n", name);
		else if (!strncmp(ValueForKey (ent, "classname"), "rotate_", 7))
		{
			int	start;
			int	end;

			GetVectorForKey (ent, "origin", org);

			start = dmodels[ i ].firstface;
			end = start + dmodels[ i ].numfaces;

			if (start >= numfaces || end > numfaces)
				Error ("FindFaceOffsets: numfaces (%d) exceeded, start=%d, end=%d", numfaces, start, end);

			for( j = start; j < end; j++ )
			{
				faceoffset[ j ][ 0 ] = org[ 0 ];
				faceoffset[ j ][ 1 ] = org[ 1 ];
				faceoffset[ j ][ 2 ] = org[ 2 ];
			}
		}
	}
}

/*
=============
LightWorld
=============
*/
void LightWorld (void)
{
	// Allocate max lightdata
	if (dlightdata != NULL)
		free (dlightdata);

	dlightdata = malloc (MAX_MAP_LIGHTING);
	lightdatasize = MAX_MAP_LIGHTING;

	filebase = file_p = dlightdata;
	file_end = filebase + MAX_MAP_LIGHTING;

	RunThreadsOn (LightThread);

	lightdatasize = file_p - filebase;

	logprintf ("lightdatasize: %s\n", PrtSize(lightdatasize));
}

/*
==============
PrintOptions
==============
*/
void PrintOptions(void)
{
	logprintf ("Light performs light processing of Quake .BSP files\n\n");
	logprintf ("light [options] bspfile\n\n");
	logprintf ("Options:\n");
	logprintf ("   -fast [n]        Enable fast lighting (lower quality, default 2)\n");
	logprintf ("   -soft [n]        Enable soft lighting (reduce jagged shadows)\n");
	logprintf ("   -softdist [n]    Distance tolerance for lights behind surface (default 3)\n");
	logprintf ("   -extra           Enable extra 2x2 sampling for higher quality\n");
	logprintf ("   -extra4          Enable extra 4x4 sampling for even higher quality\n");
	logprintf ("   -dist [n]        Set fade distance, higher is darker (default 1.0)\n");
	logprintf ("   -range [n]       Set brightness range, higher is brighter (default 0.5)\n");
	logprintf ("   -globrange       Enable global range (range affects global light)\n");
	logprintf ("   -noglobrange     Disable global range (use with arghlite/tyrlite)\n");
	logprintf ("   -light [n]       Set minimum light level (default 0)\n");
	logprintf ("   -maxlight [n]    Set maximum light level (default disabled)\n");
	logprintf ("   -nolight         Disable light entities, only global light remains\n");
	logprintf ("   -srclight        Disable all unsourced light entities\n");
	logprintf ("   -sunlight [n]    Set sunlight level (default 0)\n");
	logprintf ("   -sunlight2 [n]   Set 2nd sunlight (outdoor minlight) level (default 0)\n");
	logprintf ("   -sunlight3 [n]   Same as sunlight2 + set Shadow sensitivity to 0.4\n");
	logprintf ("   -sunmangle [y,p] Set sun direction, y=yaw (0 to 360),\n");
	logprintf ("                    p=pitch (90 to -90) (default (0,-90)\n");
	logprintf ("   -nowarnings      Disable repetitive warnings\n");
	logprintf ("   -rate [s,p,l,t]  Control extended progress update rate and format,\n");
	logprintf ("                    s=seconds, p=percent, l=line, t=total (default 10,1.0,1,1)\n");
	logprintf ("   -barpercent      Simplified and weighted bargraph progress information\n");
	logprintf ("   -numpercent      Simplified and weighted numerical progress information\n");
	logprintf ("   -oldlight        Disable entity wait/delay/mangle/antilight features\n");
	logprintf ("   -tyrlite         Enable TyrLite mode\n");
	logprintf ("   -tyrlite95       Enable TyrLite95 mode\n");
	logprintf ("   -arghlite        Enable ArghLite mode\n");
	logprintf ("   -addmin          Enable additive minlight\n");
	logprintf ("   -iklite          Enable IKLite mode\n");
	logprintf ("   -ikangle         Enable IKLite angle sensitivity\n");
	logprintf ("   -anglesense [n]  Set angle sensitivity (default 0.5)\n");
	logprintf ("   -shadowsense [n] Set angle sensitivity for 2nd sunlight (default 0.0)\n");
	logprintf ("   -gate [n]        Set Fade Gate (limit attenuated lights) (default 0.0)\n");
	logprintf ("   -dlx             Enable limited LightDLX mode\n");
	logprintf ("   -kinn            Translate delay 2 lights into delay 5\n");
	logprintf ("   -solidsky        Enable solid sky brushes\n");
	logprintf ("   -unsup           Print details about unsupported keys\n");
	logprintf ("   -detect          Print detected keys summary\n");
	logprintf ("   -noskill         Disable fuzzy skill target check\n");
	logprintf ("   -noflash         Disable flashing lights\n");
	logprintf ("   -noanti          Disable antilights\n");
	logprintf ("   -lightcap [n]    Set maximum light entity intensity\n");
	logprintf ("   -onlyents        Only process switchable light entities\n");
	logprintf ("   -nowrite         Disable bsp write\n");
	logprintf ("   -norev           Prevent reverse entity key order\n");
	logprintf ("   -etp             Enable Enhanced Texture Positioning support\n");
	logprintf ("   -priority [n]    Set thread priority 0-2 (below/normal/above, default 1)\n");
	logprintf ("   -oldhformat      Enable hour format HH:MM:SS instead of HHh MMm\n");
	logprintf ("   bspfile          .BSP file to process\n");

	fclose(logfile);

	exit(1);
}

/*
========
main

light modelfile
========
*/
int main (int argc, char **argv)
{
	int	 i, ModeCnt = 0, Val;
	char	 source[1024], *Option, *NextOption;
	float	 FVal;
	qboolean NoGlobRange = false;

        logfile = fopen (LOGFILENAME, "w");
	logprintf ("----- Light 1.43 ----\n\nmodified by Bengt Jardrup\nLinux port by gb\n\n");

	for (i=1 ; i<argc ; i++)
	{
		Option = argv[i];
		NextOption = i + 1 < argc ? argv[i + 1] : NULL;

		if (Option[0] != '-')
			break;

		++Option;

		if (!strcasecmp(Option, "fast"))
		{
			FastLight = 2;

			if (NextOption != NULL && isdigit(NextOption[0]) && i + 2 < argc)
			{
				Val = atoi (NextOption);
				i++;

				if (Val > 2)
					FastLight = Val;
			}

			logprintf ("Fast light %d enabled\n", FastLight);
		}
		else if (!strcasecmp(Option, "soft"))
		{
			SoftLight = 0;

			if (NextOption != NULL && isdigit(NextOption[0]) && i + 2 < argc)
			{
				SoftLight = atoi (NextOption);
				i++;
			}
		}
		else if (!strcasecmp(Option, "softdist"))
		{
			SoftDist = GetArgument (Option, NextOption);
			i++;
		}
		else if (!strcasecmp(Option, "extra") || !strcasecmp(Option, "extra4"))
		{
			OverSample = Option[5] == '4' ? 4 : 2;
			logprintf ("Extra %dx%d sampling enabled\n", OverSample, OverSample);
		}
		else if (!strcasecmp(Option, "dist"))
		{
			scaledist = GetFloatArgument (Option, NextOption);
			i++;
		}
		else if (!strcasecmp(Option, "range"))
		{
			rangescale = GetFloatArgument (Option, NextOption);
			i++;
		}
		else if (!strcasecmp(Option, "globrange"))
		{
			GlobRange = true;
			logprintf ("Global range enabled\n");
		}
		else if (!strcasecmp(Option, "noglobrange"))
			NoGlobRange = true;
		else if (!strcasecmp(Option, "light"))
		{
			worldminlight = GetArgument (Option, NextOption);
			i++;
		}
		else if (!strcasecmp(Option, "maxlight"))
		{
			worldmaxlight = GetArgument (Option, NextOption);
			i++;
		}
		else if (!strcasecmp(Option, "nolight"))
		{
			NoLight = true;
			logprintf ("Light entities disabled\n");
		}
		else if (!strcasecmp(Option, "srclight"))
		{
			SrcLight = true;
			logprintf ("Unsourced light entities disabled\n");
		}
		else if (!strcasecmp(Option, "sunlight"))
		{
			SunLight[0] = GetArgument (Option, NextOption);
			i++;
		}
		else if (!strcasecmp(Option, "sunlight2"))
		{
			SunLight[1] = GetArgument (Option, NextOption);
			i++;
		}
		else if (!strcasecmp(Option, "sunlight3"))
		{
			SunLight[1] = GetArgument (Option, NextOption);
			
			if (ShadowSense == -1)
				ShadowSense = SHADOWSENSE;

			i++;
		}
		else if (!strcasecmp(Option, "sunmangle"))
		{
			ChkArgument (Option, NextOption);

			if (sscanf (NextOption, "%d,%d", &SunMangleVal[0], &SunMangleVal[1]) != 2)
				Error ("Missing arguments for '%s'", Option);

			SunMangleVal[2] = 0;
			i++;
		}
		else if (!strcasecmp(Option, "nowarnings"))
			NoWarnings = true;
		else if (!strcasecmp(Option, "rate"))
		{
			ChkArgument (Option, NextOption);

			if (sscanf (NextOption, "%d,%f,%d,%d", &SecRate, &FVal, &NewLine, &TotTime) > 1)
				PercRate = (FVal + 0.05) * 10; // Fix roundoff

			AutoRate = false;
			i++;
		}
		else if (!strcasecmp(Option, "barpercent"))
		{
			SimpPercent = true;
			NumPercent = false;
		}
		else if (!strcasecmp(Option, "numpercent"))
		{
			SimpPercent = true;
			NumPercent = true;
		}
		else if (!strcasecmp(Option, "oldlight"))
		{
			OldLight = NoAnti = true;
			logprintf ("Oldlight mode enabled\n");
			++ModeCnt;
		}
		else if (!strcasecmp(Option, "tyrlite"))
		{
			TyrLiteMode = GlobRange = true;
			logprintf ("TyrLite mode enabled\n");
			++ModeCnt;
		}
		else if (!strcasecmp(Option, "tyrlite95"))
		{
			TyrLite95Mode = TyrLiteMode = GlobRange = true;
			logprintf ("TyrLite95 mode enabled\n");
			++ModeCnt;
		}
		else if (!strcasecmp(Option, "arghlite"))
		{
			ArghLiteMode = GlobRange = AddMinLight = true;
			logprintf ("ArghLite mode enabled\n");
			++ModeCnt;
		}
		else if (!strcasecmp(Option, "addmin"))
		{
			AddMinLight = true;
			logprintf ("Additive minlight enabled\n");
		}
		else if (!strcasecmp(Option, "iklite"))
		{
			IKLiteMode = IKAngle = OldLight = NoAnti = true;
			logprintf ("IKLite mode enabled\n");
			++ModeCnt;
		}
		else if (!strcasecmp(Option, "ikangle"))
		{
			IKAngle = true;
			logprintf ("IKLite angle sensitivity enabled\n");
		}
		else if (!strcasecmp(Option, "anglesense"))
		{
			FVal = GetFloatArgument (Option, NextOption);
			i++;

			if (FVal >= 0 && FVal <= 1 && FVal != scalecos)
				logprintf ("Angle sensitivity %g set\n", scalecos = FVal);
		}
		else if (!strcasecmp(Option, "shadowsense"))
		{
			FVal = GetFloatArgument (Option, NextOption);
			i++;

			if (FVal >= 0 && FVal <= 1)
				ShadowSense = FVal;
		}
		else if (!strcasecmp(Option, "gate"))
		{
			FVal = GetFloatArgument (Option, NextOption);
			i++;

			if (FVal >= 0 && FVal != GateVal)
				logprintf ("Fade Gate %g set\n", GateVal = FVal);
		}
		else if (!strcasecmp(Option, "dlx"))
		{
			LightDLXMode = OldLight = true;
			logprintf ("LightDLX mode enabled\n");
			++ModeCnt;
		}
		else if (!strcasecmp(Option, "kinn"))
		{
			KinnDelay = true;
			logprintf ("Kinn translation enabled\n");
		}
		else if (!strcasecmp(Option, "solidsky"))
		{
			SolidSky = true;
			logprintf ("Solid sky brushes enabled\n");
		}
		else if (!strcasecmp(Option, "unsup"))
			UnsupDetails = DetectKeys = true;
		else if (!strcasecmp(Option, "detect"))
			DetectKeys = true;
		else if (!strcasecmp(Option, "noskill"))
			NoSkillChk = true;
		else if (!strcasecmp(Option, "noflash"))
			NoFlash = true;
		else if (!strcasecmp(Option, "noanti"))
			NoAnti = NoAntiOption = true;
		else if (!strcasecmp(Option, "lightcap"))
		{
			Val = GetArgument (Option, NextOption);
			i++;

			if (Val >= 0 && Val != LightCap)
				logprintf ("LightCap %d enabled\n", LightCap = Val);
		}
		else if (!strcasecmp(Option, "onlyents"))
			OnlyEnts = true;
		else if (!strcasecmp(Option, "nowrite"))
			NoWrite = true;
		else if (!strcasecmp(Option, "norev"))
			NoReverse = true;
		else if (!strcasecmp(Option, "etp"))
		{
			EnhancedTP = true;
			logprintf ("Enhanced Texture Positioning enabled\n");
		}
		else if (!strcasecmp(Option, "priority"))
		{
			SetQPriority (GetArgument(Option, NextOption));
			i++;
		}
		else if (!strcasecmp(Option, "oldhformat"))
			OldHFormat = true;
		else if (!strcasecmp(Option, "?") || !strcasecmp(Option, "help"))
			PrintOptions ();
		else
			Error ("Unknown option '%s'", Option);
	}

	if (i != argc - 1)
		PrintOptions ();

	if (GlobRange && NoGlobRange)
	{
		GlobRange = false;
		logprintf ("Global range disabled\n");
	}

	if (ModeCnt > 1)
		Error ("Only one emulation mode allowed");

	if (SoftLight >= 0)
	{
		if (SoftLight == 0)
			SoftLight = OverSample / 2 + 1; // Auto mode

		logprintf ("Soft light %d enabled\n", SoftLight);

		if (SoftDist != SOFTDIST)
			logprintf ("Soft distance %d set\n", SoftDist);
	}

	if (SoftLight > 0)
	{
		// Soft option is in charge
		SingleDist = -SoftDist;
		SkyDist = -SOFTDIST;
	}
	else
	{
		if (TyrLite95Mode)
		{
			SingleDist = 0;
			SkyDist = -ANGLE_EPSILON;
		}
		else
			SingleDist = SkyDist = 1; // Classic (no) soft distance
	}

	if (OverSample <= 2)
	{
		// Set compatibility modes
		GenCompatible = OldLight || TyrLiteMode || ArghLiteMode || IKLiteMode || LightDLXMode;
		ATCompatible = TyrLiteMode || ArghLiteMode;
		TyrCompatible = TyrLiteMode;
	}

	InitThreads ();

	if (!OnlyEnts && !NoWrite)
		start = I_FloatTime ();

	strcpy (source, argv[i]);
	StripExtension (source);
	DefaultExtension (source, ".bsp");

	if (!NoWrite)
		WriteChk (source);

	logprintf ("File: %s\n", source);

	LoadBSPFile (source);	// gb, was (source, true)

	LoadEntities ();

	if (OnlyEnts)
		logprintf ("Updating entities lump...\n");
	else if (!NoWrite)
	{
		MakeTnodes (&dmodels[0]);

		FindFaceOffsets ();
		LightWorld ();
	}

	if (!NoWrite)
	{
		WriteEntitiesToString ();
		WriteBSPFile (source);
	}

	PrintFinish ();

	return 0;
}

